/* SIMPLE - Simple Is a Macro Processing Language Element */
/* Copyright (c) 1998 David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This file takes care of expanding tokens; it is in many ways the
 * central module.
 */

#include <stdio.h>
#include <stdlib.h>

#include "simple.h"
#include "token.h"

void
shipout(token_list *tl,token tk)
     /* Append the token tk to the token list tl.  If tl==NULL then
      * output the token. */
{
  if (!tl) outputtoken(tk);
  else appendtoken(tl,tk);
}

static token_list input_list={NULL,0,0};
/* This contains tokens that were generated by macro expansion and are
 * to be reprocessed. */

token
shipin(void)
     /* Read a token from the input list or the input file if the
      * input list is empty. */
{
  token tk;
  if (isempty(&input_list))
    return inputtoken();
  else {
    tk=firsttoken(&input_list);
    removefirst(&input_list);
    return tk;
  }
}

void expandtoken(token_list *tl,token tk);

void
expandargs(int *argc,token_list **argv)
     /* Read in the arguments to a macro and expand them.  Return in
      * argc the number of arguments and in argv a malloc()ated array
      * of token lists corresponding to the various arguments (after
      * expansion of course). */
{
  token tk;
  *argc=1;
  *argv=malloc(sizeof(token_list));
  while (1) {
    makeemptylist(*argv+(*argc-1));
    while (1) {
      tk=shipin();
      if (tk==EOF)
	fatal(0,"Unterminated command.");
      if ((tk==END_COMMAND)||(tk==NEXT_PARAM))
	break;
      expandtoken(*argv+(*argc-1),tk);
    }
    if (tk==END_COMMAND)
      break;
    (*argc)++;
    if ((*argv=realloc(*argv,*argc*sizeof(token_list)))==NULL)
      fatal(0,"Out of memory!");
  }
}

void
callmacro(token_list *expansion,int argc,token_list *argv)
     /* Call a (user or builtin) macro, with argc arguments given by
      * argv (of course, argv[0] is the macro name itself), and put
      * its expansion in expansion. */
{
  char *name;
  command cmd;
#if 0
  if (argc<1)
    fatal(1,"argc<1");
#else
  if (argc<1)
    return;
  /* Void call is not an error; it is just ignored. */
#endif
  name=tokenlisttostring(argv);
  cmd=identifycommand(name);
  free(name);
  if (cmd<0)
    callbuiltin(expansion,cmd,argc,argv);
  else
    calluser(expansion,cmd,argc,argv);
}

void
expandmacro(void)
     /* Read the arguments to a macro and expand it, putting the
      * expansion back on the input list. */
{
  int argc;
  token_list *argv;
  token_list expansion;
  expandargs(&argc,&argv);
  makeemptylist(&expansion);
  callmacro(&expansion,argc,argv);
  for (;argc>0;argc--) freelist(argv+argc-1);
  free(argv);
  prependlist(&input_list,&expansion);
  freelist(&expansion);
}

void
expandtoken(token_list *tl,token tk)
     /* Expand the token tk to the token list tl.  If tl==NULL, expand
      * to the output. */
{
  int quotelevel;
  if ((tk==EOF)||(tk<FIRST_SPECIAL))
    shipout(tl,tk);
  else if (tk==QUOTE_NEXT) {
    tk=shipin();
    shipout(tl,tk);
  } else if (tk==OPEN_QUOTE) {
    quotelevel=1;
    while (quotelevel>=1) {
      tk=shipin();
      if (tk==OPEN_QUOTE)
	quotelevel++;
      else if (tk==CLOSE_QUOTE)
	quotelevel--;
      else if (tk==QUOTE_NEXT)
	tk=shipin();
      if (quotelevel>=1)
	shipout(tl,tk);
    }
  } else if (tk==BEGIN_COMMAND)
    expandmacro();
  /* Note that macro expansion goes to the INPUT LIST not to the tl
   * token list, so it will be reread. */
  else if (tk==AT_SIGN)
    shipout(tl,tk);
  /* At sign gets shipped out so we don't have to put quote characters
   * around every single damn macro. Of course, if it finds its way to
   * the output, it gets kicked out at that point. */
  else
    fatal(0,"Misplaced special token %d.",tk);
}

void
expandinput(void)
     /* This is the main program loop. */
     /* Read the input and expand tokens as we go. */
{
  while (1)
    expandtoken(NULL,shipin());
}
